fn sqr x = x*x

fn getMaxReleaseAsString=(
	local ver=maxversion()
	local result
	
	if (ver[1]>=10000) then (
		result=((((ceil (ver[1]/1000.0))-2+2000) as Integer) as String)
	)
	else (
		result="R"+(((ver[1]/1000) as Integer) as String)
	)
	result
)

fn getVRayInstallPath release=(
		local regKeyName="SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall"
		local regKey
		local subKey
		local result
		
		try (
			if (registry.openKey HKEY_LOCAL_MACHINE regKeyName accessRights:#readOnly key:&regKey) do (
				local subKeyNames=#()
				registry.getSubKeyNames regKey names: &subKeyNames
				
				local i=1
				for i=1 to (subKeyNames.count) do (
					if (matchPattern subKeyNames[i] pattern:("V-Ray for 3dsmax "+release+" for*")) do (
						registry.openKey regKey subKeyNames[i] accessRights:#readOnly key:&subKey
						registry.queryValue subKey "UninstallString" value:&result
						
						registry.closeKey subKey
					)
				)
				
				registry.closeKey regKey
			)
		)
		catch (
			if (subKey!=undefined) do (registry.closeKey subKey)
			if (regKey!=undefined) do (registry.closeKey regKey)
		)
		
		if (result!=undefined) do (
			result=substring result 2 ((findString result "\\uninstall\\wininstaller.exe")-2)
		)
		
		result
)

-- used to look up materials that have already been converted
orig_mtls=#()
new_mtls=#()

convertToVRay -- Placeholder for the function, otherwise maxscript returns error when executing
					-- the rollout's convert button event if evaluated with the function defined
					-- afterwards.

rollout vraySceneConverterRollout "V-Ray scene converter" width: 460 height: 123 (
	label label_message "All scene materials and raytraced shadows will be converted to V-Ray materials and shadows." width:400 height:26 align:#center offset:[0, 10]
	
	checkbox check_optReflections "Turn on \"trace reflections\" when converting Standard materials" checked:false width:400 align:#center offset:[0, 10]
	checkbox check_optLights "Convert mr Sun/Sky to V-Ray Sun/Sky" checked:false width:400 align:#center offset:[0, 0]
	
	button btn_convertStart "Convert" width:60 align:#right across:2 offset:[150, 0]
	button btn_convertCancel "Cancel" width:60 align:#right
	
	on btn_convertStart pressed do (
		convertToVRay()
		destroyDialog vraySceneConverterRollout
	)
	
	on btn_convertCancel pressed do (
		destroyDialog vraySceneConverterRollout
	)
)

fn convertFrom_Standard origMtl = (
	r=VRayMtl()
	r.name="vray "+origMtl.name
	
	r.texmap_bump=origMtl.bumpMap
	if (r.texmap_bump!=undefined) then (
		r.texmap_bump_on=origMtl.bumpMapEnable
		r.texmap_bump_multiplier=origMtl.bumpMapAmount
	)

	r.texmap_displacement=origMtl.displacementMap
	if (r.texmap_displacement!=undefined) then (
		r.texmap_displacement_on=origMtl.displacementMapEnable
		r.texmap_displacement_multiplier=origMtl.displacementMapAmount
	)

	r.diffuse=origMtl.diffuse
	r.texmap_diffuse=origMtl.diffuseMap
	if (r.texmap_diffuse!=undefined) then (
		r.texmap_diffuse_on=origMtl.diffuseMapEnable
		r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
	)

	r.reflection=origMtl.specular*origMtl.specularLevel*0.01
	
	K=2.0^(origMtl.glossiness*0.01*10.0)
	if (origMtl.shaderType==5) then (
		-- phong
		r.brdf_type=0
		r.reflection=r.reflection*2.0/(K+1.0)
	) else (
		-- blinn
		r.brdf_type=1
		K=K*4
		r.reflection=r.reflection*8.0/(K+1.0)
	)
	
	r.reflection_lockGlossiness=false
	r.hilight_glossiness=1.0-(1.0/(K+1.0))^(1.0/3.5)
	r.reflection_glossiness=1.0
	r.option_traceReflection=vraySceneConverterRollout.check_optReflections.checked
	
	r.texmap_opacity=origMtl.opacityMap
	if (r.texmap_opacity!=undefined) then (
		r.texmap_opacity_on=origMtl.opacityMapEnable
		r.texmap_opacity_multiplier=origMtl.opacityMapAmount
	)
	
	r
)

fn convertFrom_Raytrace origMtl = (
	r=VRayMtl()
	r.name="vray "+origMtl.name

	r.texmap_bump=origMtl.bumpMap
	if (r.texmap_bump!=undefined) then (
		r.texmap_bump_on=origMtl.bumpMapEnable
		r.texmap_bump_multiplier=origMtl.bumpMapAmount
	)

	r.texmap_displacement=origMtl.displacementMap
	if (r.texmap_displacement!=undefined) then (
		r.texmap_displacement_on=origMtl.displacementMapEnable
		r.texmap_displacement_multiplier=origMtl.displacementMapAmount
	)

	r.diffuse=origMtl.diffuse
	r.texmap_diffuse=origMtl.diffuseMap
	if (r.texmap_diffuse!=undefined) then (
		r.texmap_diffuse_on=origMtl.diffuseMapEnable
		r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
	)

	r.reflection=origMtl.spec__color*origMtl.specular_level*0.01*0.5
	
	K=2.0^(origMtl.glossiness*0.01*10.0)
	if (origMtl.shaderType==1) then (
		-- phong
		r.brdf_type=0
		r.reflection=r.reflection*2.0/(K+1.0)
	) else (
		-- blinn
		r.brdf_type=1
		K=K*4
		r.reflection=r.reflection*8.0/(K+1.0)
	)

	if (origMtl.reflect.r>r.reflection.r) then r.reflection.r=origMtl.reflect.r
	if (origMtl.reflect.g>r.reflection.g) then r.reflection.g=origMtl.reflect.g
	if (origMtl.reflect.b>r.reflection.b) then r.reflection.b=origMtl.reflect.b
	
	r.reflection_lockGlossiness=false
	r.hilight_glossiness=1.0-(1.0/(K+1.0))^(1.0/3.5)
	
	if ((origMtl.reflect.r+origMtl.reflect.g+origMtl.reflect.b)/3.0<1e-6) then r.option_traceReflection=false
	
	r.texmap_reflection=origMtl.reflectionMap
	if (r.texmap_reflection!=undefined) then (
		r.texmap_reflection_on=origMtl.reflectionMapEnable
		r.texmap_reflection_multiplier=origMtl.reflectionMapAmount
	)
	
	r.refraction=origMtl.transparecy
	r.refraction_ior=origMtl.Index_of_Refraction

	r.texmap_refraction=origMtl.transparencyMap
	if (r.texmap_refraction!=undefined) then (
		r.texmap_refraction_on=origMtl.transparencyMapEnable
		r.texmap_refraction_multiplier=origMtl.transparencyMapAmount
	)

	r
)

fn convertFrom_fR_Advanced origMtl = (
	r=VRayMtl()
	r.name="vray "+origMtl.name
	
	-- diffuse
	r.diffuse=origMtl.diffuseColor
	r.texmap_diffuse=origMtl.diffuseMap
	r.texmap_diffuse_on=origMtl.useDiffuseMap
	r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
	
	-- bump map
	r.texmap_bump=origMtl.bumpMap
	r.texmap_bump_on=origMtl.useBumpMap
	r.texmap_bump_multiplier=origMtl.bumpMapAmount
	
	-- displacement
	r.texmap_displacement=origMtl.MTDMap
	r.texmap_displacement_on=origMtl.useMTDMap
	r.texmap_displacement_multiplier=origMtl.MTDMapAmount

	-- reflection
	r.reflection=origMtl.reflectFilter
	r.texmap_reflection=origMtl.reflectMap
	r.texmap_reflection_on=origMtl.useReflectMap
	r.texmap_reflection_multiplier=origMtl.reflectMapAmount
	
	r.reflection_glossiness=1.0-sqr (origMtl.reflectGlossy*0.01)
	r.texmap_reflectionGlossiness=origMtl.reflectBlurryMap
	r.texmap_reflectionGlossiness_on=origMtl.useReflectBlurryMap
	r.texmap_reflectionGlossiness_multiplier=origMtl.reflectBlurryMapAmount
	
	r.reflection_subdivs=sqrt origMtl.reflectSamples
	r.reflection_fresnel=origMtl.useFresnel
	r.reflection_useInterpolation=origMtl.useReflBlur
	
	-- refraction
	r.refraction=origMtl.refractFilter
	r.texmap_refraction=origMtl.refractMap
	r.texmap_refraction_on=origMtl.useRefractMap
	r.texmap_refraction_multiplier=origMtl.refractMapAmount
	
	r.refraction_ior=origMtl.refractIOR
	r.texmap_refractionIOR=origMtl.refractIORMap
	r.texmap_refractionIOR_on=origMtl.useRefractIORMap
	r.texmap_refractionIOR_multiplier=origMtl.refractIORMapAmount
	
	r.refraction_glossiness=1.0-sqr (origMtl.refractGlossy*0.01)
	r.texmap_refractionGlossiness=origMtl.refractBlurryMap
	r.texmap_refractionGlossiness_on=origMtl.useRefractBlurryMap
	r.texmap_refractionGlossiness_multiplier=origMtl.refractBlurryMapAmount
	
	r.refraction_subdivs=sqrt origMtl.refractSamples
	r.refraction_useInterpolation=origMtl.useRefrBlur
	
	r
)

fn convertFrom_fR_Metal origMtl = (
	r=VRayMtl()
	r.name="vray "+origMtl.name
	
	r.diffuse=origMtl.diffuseColor
	r.texmap_diffuse=origMtl.diffuseMap
	r.texmap_diffuse_on=origMtl.diffuseMapEnable
	r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount
	
	r.reflection=origMtl.reflectColor
	r.texmap_reflection=origMtl.reflectLevelMap
	r.texmap_reflection_on=origMtl.reflectLevelMapEnable
	r.texmap_reflection_multiplier=origMtl.reflectLevelMapAmount
	
	r
)

fn convertFrom_fR_Glass origMtl = (
	r=VRayMtl()
	r.name="vray "+origMtl.name
	
	r.diffuse=origMtl.diffuseColor
	r.texmap_diffuse=origMtl.diffuseMap
	r.texmap_diffuse_on=origMtl.diffuseMapEnable
	r.texmap_diffuse_multiplier=origMtl.diffuseMapAmount

	r.reflection=origMtl.reflectColor
	r.reflection_fresnel=origMtl.useFresnel
	r.texmap_reflection=origMtl.reflectMap
	r.texmap_reflection_on=origMtl.reflectMapEnable
	r.texmap_reflection_multiplier=origMtl.reflectMapAmount
	
	r.refraction=origMtl.refractColor
	r.refraction_ior=origMtl.refractIOR
	r.texmap_refraction=origMtl.refractMap
	r.texmap_refraction_on=origMtl.refractMapEnable
	r.texmap_refraction_multiplier=origMtl.refractMapAmount
	
	r.refraction_affectShadows=true
	r.refraction_affectAlpha=1
	
	r
)

fn convertFromBlend orig_mtl = (
	r=VRayBlendMtl()
	r.name="vray "+orig_mtl.name
	
	r.baseMtl=orig_mtl.map1
	r.coatMtl[1]=orig_mtl.map2
	r.blend[1]=(orig_mtl.mixAmount*0.01)*(color 255 255 255)
	if orig_mtl.maskEnabled then r.texmap_blend[1]=orig_mtl.mask
	
	r
)

fn convertFromShellac orig_mtl = (
	r=VRayBlendMtl()
	r.name="vray "+orig_mtl.name
	r.additiveMode=true
	
	r.baseMtl=orig_mtl.shellacMtl1
	r.coatMtl[1]=orig_mtl.shellacMtl2
	r.blend[1]=(orig_mtl.shellacColorBlend*0.01)*(color 255 255 255)
	
	r
)

fn convertFromDoubleSided orig_mtl = (
	r=VRay2SidedMtl()
	r.name="vray "+orig_mtl.name

	r.frontMtl=orig_mtl.material1
	r.backMtl=orig_mtl.material2
	r.translucency=orig_mtl.translucency*0.01*(color 255 255 255)
	r.backMtlOn=orig_mtl.map2Enabled
	
	r
)

fn convertFrom_Architectural orig_mtl = (
	r=VRayMtl()
	r.name="vray "+orig_mtl.name

	-- diffuse component
	r.diffuse=orig_mtl.diffuse
	r.texmap_diffuse=orig_mtl.diffuseMap
	if (r.texmap_diffuse!=undefined) then (
		r.texmap_diffuse_on=orig_mtl.diffuseMapEnable
		r.texmap_diffuse_multiplier=orig_mtl.diffuseMapAmount
	)

	-- specular component
	r.reflection=(color 255 255 255)
	r.texmap_reflection=orig_mtl.shininessMap
	r.reflection_fresnel=true
	r.hilight_glossiness=(orig_mtl.shininess*0.01)*0.25+0.75
	r.reflection_glossiness=1.0
	r.reflection_lockGlossiness=false

	-- refraction
	r.refraction=orig_mtl.transparency*0.01*(color 255 255 255)
	r.refraction_ior=orig_mtl.ior
	
	-- if there is luminance, wrap the material into a VRayBlendMtl and add a VRayLightMtl on top of it
	if orig_mtl.luminance>0 then (
		local l=VRayLightMtl()
		l.color=orig_mtl.diffuse
		l.multiplier=orig_mtl.luminance
		l.opacity_texmap=VRayColor()
		l.opacity_texmap.color=(color 0 0 0)
		
		local m=VRayBlendMtl()
		m.baseMtl=r
		m.coatMtl[1]=l
		m.blend[1]=(color 255 255 255)
		
		r=m
	)
	
	-- if there is translucency, wrap the material into a VRay2SidedMtl
	if orig_mtl.translucency>0 then (
		local m=VRay2SidedMtl()
		m.frontMtl=r
		m.translucency=orig_mtl.translucency*0.01*(color 255 255 255)
		r=m
	)
	
	r
)

fn convertFrom_ArchAndDesign orig_mtl = (
	r=VRayMtl()
	r.name="vray "+orig_mtl.name
	
	-- diffuse component
	r.diffuse=orig_mtl.diff_color*orig_mtl.diff_weight
	r.texmap_diffuse=orig_mtl.diff_color_map
	if (r.texmap_diffuse!=undefined) then (
		r.diffuse=(color 0 0 0)
		r.texmap_diffuse_on=orig_mtl.diff_color_map_on
		r.texmap_diffuse_multiplier=orig_mtl.diff_weight*100
	)
	
	-- reflection
	r.reflection=orig_mtl.refl_color*orig_mtl.refl_weight
	r.texmap_reflection=orig_mtl.refl_color_map
	if (r.texmap_reflection!=undefined) then (
		r.reflection=(color 0 0 0)
		r.texmap_reflection_on=orig_mtl.refl_color_map_on
		r.texmap_reflection_multiplier=orig_mtl.refl_weight*100
	)
	
	r.reflection_glossiness=orig_mtl.refl_gloss
	r.texmap_reflectionGlossiness=orig_mtl.refl_gloss_map
	if (r.texmap_reflectionGlossiness!=undefined) then (
		r.reflection_glossiness=0.0
		r.texmap_reflectionGlossiness_on=orig_mtl.refl_gloss_map_on
	)
	
	-- Fresnel options
	if (orig_mtl.refl_func_fresnel==true) then ( -- use Fresnel
		r.reflection_fresnel=true
	) else ( -- use falloff curve
		local tcolor=Falloff()
		if (r.texmap_reflection==undefined) then (
			tcolor.color1=orig_mtl.refl_color*orig_mtl.refl_func_low
			tcolor.color2=orig_mtl.refl_color*orig_mtl.refl_func_high
		) else (
			tcolor.color1=(color 1.0 1.0 1.0)*orig_mtl.refl_func_low
			tcolor.color2=(color 1.0 1.0 1.0)*orig_mtl.refl_func_high
			
			local m=VRayCompTex()
			m.sourceA=r.texmap_reflection
			m.sourceB=tcolor
			m.operator=3
			tcolor=m
		)
		
		r.texmap_reflection=tcolor
		r.texmap_reflection_on=true
		r.texmap_reflection_multiplier=orig_mtl.refl_weight*100
		r.reflection=(color 0 0 0)
	)
	
	-- refraction/opacity
	r.refraction_ior=orig_mtl.refr_ior
	r.refraction_affectShadows=not orig_mtl.opts_do_refractive_caustics
	r.refraction_affectAlpha=(if orig_mtl.opts_propagate_alpha then 1 else 0)
	if (orig_mtl.opts_1sided==on) then ( -- opacity
		if (orig_mtl.refr_color_map!=undefined) then ( -- there is a transparency texture
			local tcolor=Output()
			tcolor.map1=orig_mtl.refr_color_map
			tcolor.output.invert=true
			tcolor.output_amount=orig_mtl.refr_weight
			
			r.texmap_opacity=tcolor
		) else ( -- no transparency texture, just color
			local c=orig_mtl.refr_color
			if (orig_mtl.refr_weight>0 and (c.r>0 or c.g>0 or c.b>0)) then ( -- there is some transparency
				local tcolor=VRayColor()
				tcolor.color=(color (1.0-orig_mtl.refr_color.r) (1.0-orig_mtl.refr_color.g) (1.0-orig_mtl.refr_color.b))
				r.texmap_opacity=tcolor
			)
		)
	) else ( -- refraction
		r.refraction=orig_mtl.refr_color*orig_mtl.refr_weight
		r.texmap_refraction=orig_mtl.refr_color_map
		if (r.texmap_refraction!=undefined) then (
			r.refraction=(color 0 0 0)
			r.texmap_refraction_on=orig_mtl.refr_color_map_on
			r.texmap_refraction_multiplier=orig_mtl.refr_weight*100
		)
	)
	
	r
)

fn isVRayCompatible mtl_classid = (
	-- native V-Ray materials
	if mtl_classid==VRayMtl then return true
	if mtl_classid==VRay2SidedMtl then return true
	if mtl_classid==VRayBlendMtl then return true
	if mtl_classid==VRayFastSSS then return true
	if mtl_classid==VRayLightMtl then return true
	if mtl_classid==VRayMtlWrapper then return true
	
	-- supported 3dsmax materials
	if mtl_classid==Multimaterial then return true
	if mtl_classid==Arch___Design__mi then return true
	
	false
)

fn createVRayMtl orig_mtl = (
	-- check if this is a material; if not, just return the value
	if ((superclassof orig_mtl)!=material) then return orig_mtl

	-- check if we have already converted this material; if yes, just return
	-- the already converted one	
	local i=findItem orig_mtls orig_mtl
	if i>0 then return new_mtls[i]

	-- recurse into sub-materials, if any
	local sub_count=getNumSubMtls orig_mtl
	if sub_count>1 then (
		for i=1 to sub_count do (
			setSubMtl orig_mtl i (createVRayMtl (getSubMtl orig_mtl i))
		)
	)

	-- get the material class id
	cls=classof orig_mtl
	
	-- if this is already a V-Ray compatible material, return the same one
	if (isVRayCompatible cls) then return orig_mtl

	-- otherwise, try to convert
	format "  %\n" orig_mtl
	local new_mtl=case cls of (
		StandardMaterial: convertFrom_Standard orig_mtl
		RaytraceMaterial: convertFrom_Raytrace orig_mtl
		Architectural: convertFrom_Architectural orig_mtl
		
		fR_Advanced: convertFrom_fR_Advanced orig_mtl
		fR_Metal: convertFrom_fR_Metal orig_mtl
		fR_Glass: convertFrom_fR_Glass orig_mtl
		
		Blend: convertFromBlend orig_mtl
		Shellac: convertFromShellac orig_mtl
		DoubleSided: convertFromDoubleSided orig_mtl
		
		-- Do not convert Arch&Design materials - V-Ray supports them directly
		-- Arch___Design__mi: convertFrom_ArchAndDesign orig_mtl
		
		default: orig_mtl
	)
	
	-- if converted, add to the list of converted materials
	if new_mtl!=orig_mtl then (
		append orig_mtls orig_mtl
		append new_mtls new_mtl
	)
	
	-- return the new material
	new_mtl
)

fn replaceMtlNode node = (
	if isValidNode node then (
		local m=createVRayMtl node.material
		node.material=m
		format "    %\n" node.name
	)
	for i=1 to node.children.count do (
		replaceMtlNode node.children[i]
	)
)

-- convert all scene materials to VRay materials
fn mtlsToVRay = (
	format "Converting scene materials...\n"
	orig_mtls=#()
	new_mtls=#()
	replaceMtlNode rootNode
	format "Converting Material Editor materials...\n"
	for i=1 to 24 do (
		local m=createVRayMtl meditmaterials[i]
		meditmaterials[i]=m
	)
	orig_mtls=#()
	new_mtls=#()
	format "Done\n"
)

fn shadowsNode node = (
	if (isValidNode node) then (
		scls=superclassof node
		if (scls==light) then (
			try (
				olds=node.shadowGenerator
				cls=classof olds
				if (cls!=shadowMap and cls!=VRayShadow) then (
					format "  %\n" node
					news=VRayShadow()
					node.shadowGenerator=news
				)
			)
			catch ()
		)
	)
	
	for i=1 to node.children.count do (
		shadowsNode node.children[i]
	)
)

-- convert all shadow types except shadow maps to VRay shadows
fn shadowsToVRay = (
	format "Converting shadows...\n"
	shadowsNode rootNode
	format "Done\n"
)

fn normalizeColor col = (
	norm=if (col.r>col.g) then (if (col.r>col.b) then col.r else col.b) else (if (col.g>col.b) then col.g else col.b)
	result=color ((col.r*255.0)/norm) ((col.g*255.0)/norm) ((col.b*255.0)/norm)
	result
)

fn replaceLightNodeVRIESLight node = (
	local newLight=CreateInstance VRayIES name:(node.name+"_VRayIESLight")
	
	if (newLight!=undefined) do (
		newLight.targeted=(node.target!=undefined)
		newLight.enabled=node.on
		newLight.color_mode=0
		newLight.color=normalizeColor node.color
		newLight.power=node.intensity*4*pi
		if (node.useMultiplier) do (newLight.power=(newLight.power*node.multiplier)/100.0)
		newLight.rotation_X=node.xRotation
		newLight.rotation_Y=node.yRotation
		newLight.rotation_Z=node.zRotation
		newLight.use_light_shape=2
		if (node.webFile!=undefined) do (newLight.ies_file=node.webFile)
		if (hasProperty newLight "override_shape") then (
			newLight.override_shape=true
			case of (
				(node.type==#Target_Point or node.type==#Free_Point): (
					newLight.shape=0
				)
				
				(node.type==#Target_Line or node.type==#Free_Line or
				node.type==#Target_Linear or node.type==#Free_Linear): (
					newLight.shape=1
					newLight.length=node.light_length
				)
				
				(node.type==#Target_Rectangle or node.type==#Free_Rectangle or
				node.type==#Target_Area or node.type==#Free_Area): (
					newLight.shape=1
					newLight.width=node.light_width
					newLight.length=node.light_length
				)
				
				(node.type==#Target_Disc or node.type==#Free_Disc): (
					newLight.shape=2
					newLight.diameter=node.light_radius*2
				)
				
				(node.type==#Target_Sphere or node.type==#Free_Sphere): (
					newLight.shape=3
					newLight.diameter=node.light_radius*2
				)
				
				(node.type==#Target_Cylinder or node.type==#Free_Cylinder): (
					newLight.shape=5
					newLight.length=node.light_length
					newLight.diameter=node.light_radius*2
				)
			)
		)
		if (node.includeList!=undefined) do (newLight.includeList=node.includeList)
		if (node.excludeList!=undefined) do (newLight.excludeList=node.excludeList)
	)
	
	node.baseObject=newLight
	newLight
)

fn replaceLight_mr_Sun lightsys = (
	if (classof lightsys.sun==mr_Sun) do (
		local mrSun_on=lightsys.sun.on
		local mrSun_skymult=lightsys.sun.skymult
		
		lightsys.sun=createInstance VRaySun
		lightsys.sun.enabled=mrSun_on
		lightsys.sun.intensity_multiplier=mrSun_skymult
		
		lightsys.sky=undefined
	)
)

fn replaceLightNode node = (
	local lightClassStandard=#(Free_Light, Target_Light,
									Free_Point, Target_Point,
									Free_Linear, Target_Linear,
									Free_Area, Target_Area,
									Free_Disc, Target_Disc,
									Free_Sphere, Target_Sphere,
									Free_Cylinder, Target_Cylinder)
	
	local newLight		
	if (isValidNode node) do (
		if ((findItem lightClassStandard (classof node))!=0) do (
			case (node.distribution) of (
				0: newLight=replaceLightNodeVRIESLight node	-- Uniform Spherical
				1: newLight=undefined						-- Spotlight
				2: newLight=replaceLightNodeVRIESLight node	-- Uniform Diffuse
				3: newLight=replaceLightNodeVRIESLight node	-- Photometric Web
			)
		)
	)
	
	if (newLight!=undefined) do (
		print (node.name)
	)
	
	newLight!=undefined
)

fn lightsToVRay = (
	format "Converting scene Lights...\n"
	local oldLights=#()
	
	-- First traverse scene for daylight systems otherwise the sun/sky lights in them will appear as standalone
	-- lights to the normal lights replace traversal.
	if (vraySceneConverterRollout.check_optLights.checked) do (
		for i=1 to $helpers.count do (
			if (classof $helpers[i]==DaylightAssemblyHead) do (
				replaceLight_mr_Sun $helpers[i]
			)
		)
		
		if ((environmentMap!=undefined) and (classof environmentMap==mr_Physical_Sky)) do (
			environmentMap=VRaySky()
		)
	)
	
	for i=1 to $lights.count do (
		if (replaceLightNode $lights[i]) do ( append oldLights $lights[i] )
	)
	-- old lights need to stay until the conversion is done, because deleteing them
	-- on the run messes nodes' order during traversal
	format "Done\n"
)

fn ConvertToVRay = (
	mtlsToVRay()
	shadowsToVRay()
	lightsToVRay()
)
convertToVRay=ConvertToVRay

-- Bitmap2VRayHDRI converter
convertB2HDRI_ConvertEXR=false
convertB2HDRI_ConverterPath=undefined
convertB2HDRI_ConvertHalf=false
convertB2HDRI_ConvertCompression="zip"
convertB2HDRI_ConvertTilesize=64
convertB2HDRI_ConvertLinear="auto"

convertB2HDRI_ConverterErrors=0

vrayBitmap2VRayHDRI
convertB2HDRI_TEXRConverterRollout
rollout convertB2HDRI_Rollout "V-Ray Bitmap to VRayHDRI texmap converter" width: 460 height: 97 (
	label label_message "All scene materials' Bitmap texmaps will be converted to VRayHDRI texmap." width:400 height:13 align:#center offset:[0, 10]
	
	checkbox check_convertEXR "Convert bitmaps to tiled EXR." checked:false width:400 align:#center offset:[0, 10]
	
	button btn_start "Convert" width:60 align:#right across:2 offset:[150, 0]
	button btn_cancel "Cancel" width:60 align:#right
	
	local cancelled=false
	local convertPath=undefined
	local convertHalf=false
	local convertCompression="zip"
	local convertTilesize=64
	local convertLinear="auto"
	
	on btn_start pressed do (
		convertB2HDRI_ConverterErrors=0
		cancelled=false
		
		if (check_convertEXR.checked) do (
			CreateDialog convertB2HDRI_TEXRConverterRollout modal:true
		)
		if (NOT cancelled) do (
			vrayBitmap2VRayHDRI exrConvert:check_convertEXR.checked exrPath:convertPath exrHalf:convertHalf exrCompress:convertCompression exrTilesize:convertTilesize exrLinear:convertLinear
			if (convertB2HDRI_ConverterErrors>0) do (
				messageBox ("There were "+(convertB2HDRI_ConverterErrors as String)+" images that did not convert successfully to tiled EXR format. The original images were left in place.") title:"Tiled EXR conversion error"
			)
		)
		
		destroyDialog convertB2HDRI_Rollout
	)
	
	on btn_cancel pressed do (
		destroyDialog convertB2HDRI_Rollout
	)
)

rollout ConvertB2HDRI_TEXRConverterRollout "Img2tiledexr tool conversion options" width: 460 height: 200 (
	label label_message "If the img2tiledexr tool was not found it should be set manually." width:400 height:13 align:#center offset:[0, 10]
	
	edittext edit_path width:394 across:2 align:#left offset:[13, 0]
	button btn_path "..." width:20 height:18 align:#right offset:[0, 0]
	
	checkbox check_half "Half - store data as 16-bit floating point numbers." checked:false width:400 align:#center offset:[0, 10]
	
	label label_compression "Compression method:" width:200 height:13 across:2 align:#center offset:[10, 2]
	dropdownlist drop_compression items:#("None", "RLE", "Zips", "Zip", "PIZ", "PXR24") selection:4 width:100 align:#center offset:[-150, 0]
	
	label label_tilesize "Tile size:" width:200 height:13 across:2 align:#center offset:[10, 2]
	dropdownlist drop_tilesize items:#("32x32", "64x64", "128x128") selection:2 width:100 align:#center offset:[-150, 0]
	
	label label_linearmode "Linear mode:" width:200 height:13 across:2 align:#center offset:[10, 2]
	dropdownlist drop_linearmode items:#("Off", "On", "Auto") selection:3 width:100 align:#center offset:[-150, 0]
	
	button btn_set "OK" width:60 align:#right across:2 offset:[150, 0]
	button btn_cancel "Cancel" width:60 align:#right
	
	on ConvertB2HDRI_TEXRConverterRollout open do (
		if (convertB2HDRI_Rollout.check_convertEXR.checked) do (
			local ver=getMaxReleaseAsString()
			local tmpPath=getVRayInstallPath ver
			if (tmpPath!=undefined) do (
				append tmpPath "\\tools\\img2tiledexr.exe"
				edit_path.text=tmpPath
				edit_path.enabled=false
				btn_path.enabled=false
			)
		)
	)
	
	on btn_path pressed do (
		local toolPath=getOpenFileName filename:edit_path.text \
								types:"Executable File (img2tiledexr.exe)|img2tiledexr.exe"
		if (toolPath!=undefined) do (
			edit_path.text=toolPath
		)
	)
	
	on btn_set pressed do (
		convertB2HDRI_Rollout.convertPath=edit_path.text
		convertB2HDRI_Rollout.convertHalf=check_half.checked
		convertB2HDRI_Rollout.convertCompression=case drop_compression.selection of (
			1:	"none"
			2:	"rle"
			3:	"zips"
			4:	"zip"
			5:	"piz"
			6:	"pxr24"
		)
		convertB2HDRI_Rollout.convertTilesize=case drop_tilesize.selection of (
			1:	32
			2:	64
			3:	128
		)
		convertB2HDRI_Rollout.convertLinear=case drop_linearmode.selection of (
			1:	"off"
			2:	"on"
			3:	"auto"
		)
		destroyDialog convertB2HDRI_TEXRConverterRollout
	)
	
	on btn_cancel pressed do (
		convertB2HDRI_Rollout.cancelled=true
		destroyDialog convertB2HDRI_TEXRConverterRollout
	)
)
convertB2HDRI_TEXRConverterRollout=ConvertB2HDRI_TEXRConverterRollout

fn convertB2HDRI_img2tiledexr imgpath=(
	if ((classof imgpath)==String) then (
		if ((stricmp (getFilenameType imgpath) ".exr")!=0) then (
			outpath=(getFilenamePath imgpath)+(getFilenameFile imgpath)+".exr"
		)
		else (
			outpath=imgpath
		)
		
		if ((classof convertB2HDRI_ConverterPath)==String) do (
			local cmdConverter=("\""+convertB2HDRI_ConverterPath+"\" \""+imgpath+"\" \""+outpath+"\"")
			if (convertB2HDRI_ConvertHalf) then (append cmdConverter " -half")
			else (append cmdConverter " -32bit")
			append cmdConverter (" -compression "+convertB2HDRI_ConvertCompression)
			append cmdConverter (" -tileSize "+(convertB2HDRI_ConvertTilesize as String))
			append cmdConverter (" -linear "+convertB2HDRI_ConvertLinear)
			
			vrayShellLaunch cmdConverter waitExec:true hideConsole:true
		)
		
		if (doesFileExist outpath) then (
			outpath
		)
		else (
			if (convertB2HDRI_ConverterErrors!=undefined) do (
				convertB2HDRI_ConverterErrors=convertB2HDRI_ConverterErrors+1
			)
			imgpath
		)
	)
	else (
		undefined
	)
)

fn convertB2HDRI_Texmap texmap=(
	if (texmap!=undefined) then (
		local tex=VRayHDRI()
		tex.name=texmap.name+"_VRay"
		if (convertB2HDRI_ConvertEXR) then (
			tex.HDRIMapName=convertB2HDRI_img2tiledexr texmap.fileName
			
			if (tex.HDRIMapName!=undefined) do (
				local isHDRI=(((stricmp (getFilenameType texmap.fileName) ".exr")==0) OR ((stricmp (getFilenameType texmap.fileName) ".hdr")==0))
				if (NOT isHDRI) do (
					local gamma=vrayGetBitmapGamma texmap
					if (gamma!=undefined) then (
						if (convertB2HDRI_ConvertLinear!="off") then (
							gamma=gamma/2.2
						)
						tex.gamma=1.0/gamma
					)
				)
			)
		)
		else (
			tex.HDRIMapName=texmap.fileName
			
			local isHDRI=(((stricmp (getFilenameType texmap.fileName) ".exr")==0) OR ((stricmp (getFilenameType texmap.fileName) ".hdr")==0))
			if (NOT isHDRI) do (
				local gamma=vrayGetBitmapGamma texmap
				if (gamma!=undefined) then (
					tex.gamma=1.0/gamma
				)
			)
		)
		tex.mapType=4
		tex.UVGen.blur=texmap.coords.blur
		tex.UVGen.mapping=texmap.coords.mapping
		tex.UVGen.mapChannel=texmap.coords.mapChannel
		tex.UVGen.mappingType=texmap.coords.mappingType
		tex.UVGen.UVW_Type=texmap.coords.UVW_Type
		tex.UVGen.U_Mirror=texmap.coords.U_Mirror
		tex.UVGen.V_Mirror=texmap.coords.V_Mirror
		tex.UVGen.U_Tile=texmap.coords.U_Tile
		tex.UVGen.V_Tile=texmap.coords.V_Tile
		tex.UVGen.showMapOnBack=texmap.coords.showMapOnBack
		tex.UVGen.Noise_On=texmap.coords.Noise_On
		tex.UVGen.Noise_Animate=texmap.coords.Noise_Animate
		tex.UVGen.realWorldScale=texmap.coords.realWorldScale
		tex.UVGen.realWorldHeight=texmap.coords.realWorldHeight
		tex.UVGen.realWorldWidth=texmap.coords.realWorldWidth
		tex.UVGen.phase=texmap.coords.phase
		tex.UVGen.U_Angle=texmap.coords.U_Angle
		tex.UVGen.V_Angle=texmap.coords.V_Angle
		tex.UVGen.W_Angle=texmap.coords.W_Angle
		tex.UVGen.Noise_Amount=texmap.coords.Noise_Amount
		tex.UVGen.Noise_Size=texmap.coords.Noise_Size
		tex.UVGen.Noise_Levels=texmap.coords.Noise_Levels
		tex.UVGen.Blur_Offset=texmap.coords.Blur_Offset
		tex.UVGen.u_tiling=texmap.coords.u_tiling
		tex.UVGen.v_tiling=texmap.coords.v_tiling
		tex.UVGen.u_offset=texmap.coords.u_offset
		tex.UVGen.v_offset=texmap.coords.v_offset
		
		tex
	)
	else (
		undefined
	)
)

fn convertB2HDRI_SubTex texmap = (
	if (texmap!=undefined) then (
		if ((classof texmap)==Bitmaptexture) then (
			local tex=convertB2HDRI_Texmap texmap
			tex
		)
		else (
			for i = 1 to (getNumSubTexmaps texmap) do (
				local tex=(convertB2HDRI_SubTex (getSubTexmap texmap i))
				if (tex!=undefined) do (
					setSubTexmap texmap i tex
				)
			)
			undefined
		)
	)
	else (
		undefined
	)
)

fn convertB2HDRI_SubMtl material = (
	if (material!=undefined) do (
		for i=1 to (getNumSubMtls material) do (
			convertB2HDRI_SubMtl (getSubMtl material i)
		)
		
		for i=1 to (getNumSubTexmaps material) do (
			local tex=(convertB2HDRI_SubTex (getSubTexmap material i))
			if (tex!=undefined) do (setSubTexmap material i tex)
		)
	)
)

fn convertB2HDRI_NodeMtl node =(
	if isValidNode node then (
		convertB2HDRI_SubMtl node.material
	)
	for i=1 to node.children.count do (
		convertB2HDRI_NodeMtl node.children[i]
	)
)

fn VrayBitmap2VRayHDRI exrConvert: exrPath: exrHalf: exrCompress: exrTilesize: exrLinear: =(
	convertB2HDRI_ConvertEXR=false
	if (exrConvert!=unsupplied) do (
		if ((classof exrConvert)!=BooleanClass) do (print "Expected Boolean for exrConvert:")
		if (exrConvert) do (
			if (exrPath!=unsupplied) then (
				if ((classof exrPath)!=String) do (print "Expected String for exrPath:")
				convertB2HDRI_ConverterPath=exrPath
			)
			else (
				if (convertB2HDRI_ConverterPath==undefined) do (
					local ver=getMaxReleaseAsString()
					convertB2HDRI_ConverterPath=getVRayInstallPath ver
					if (convertB2HDRI_ConverterPath!=undefined) do (append convertB2HDRI_ConverterPath "\\tools\\img2tiledexr.exe")
				)
			)
		)
		convertB2HDRI_ConvertEXR=exrConvert
	)
	
	if (exrHalf!=unsupplied) do (
		if ((classof exrHalf)!=BooleanClass) do (print "Expected Boolean for exrHalf:")
		convertB2HDRI_ConvertHalf=exrHalf
	)
	if (exrCompress!=unsupplied) do (
		if ((classof exrCompress)!=String) do (print "Expected String for exrCompress:")
		convertB2HDRI_ConvertCompression=exrCompress
	)
	if (exrTilesize!=unsupplied) do (
		if ((classof exrTilesize)!=Integer) do (print "Expected Integer for exrTilesize:")
		convertB2HDRI_ConvertTilesize=exrTilesize
	)
	if (exrLinear!=unsupplied) do (
		if ((classof exrLinear)!=String) do (print "Expected String for exrLinear:")
		convertB2HDRI_ConvertLinear=exrLinear
	)
	
	convertB2HDRI_NodeMtl rootNode
	
	for i=1 to meditMaterials.count do (
		local m=meditMaterials[i]
		if (m.superClassID==3072) then ( -- This is a material
			convertB2HDRI_SubMtl m
		) else ( -- This is a texture
			m=convertB2HDRI_SubTex m
			if (m!=undefined) then meditMaterials[i]=m
		)
		updateMTLInMedit meditMaterials[i]
	)
)
vrayBitmap2VRayHDRI=VrayBitmap2VRayHDRI

fn registerMenuItems = (
	local quadMenu=menuMan.getViewportRightClickMenu #nonePressed
	local menu=quadMenu.getMenu 1
	
	-- register old menus
	if menuMan.registerMenuContext 0x763abd78 then (
		local vrayProperties=menuMan.createActionItem "VRayProperties" "VRay"
		menu.addItem vrayProperties -1
		
		local vrayConverter=menuMan.createActionItem "VRaySceneConverter" "VRay"
		menu.addItem vrayConverter -1
	)
	-- register new menus
	if menuMan.registerMenuContext 0x3f523d41 then (
		local vrayMeshExport=menuMan.createActionItem "VRayMeshExport" "VRay"
		menu.addItem vrayMeshExport -1
	)
	-- register newer menus
	if menuMan.registerMenuContext 0x45e60e47 then (
		local vrayShowVFB=menuMan.createActionItem "VRayShowVFB" "VRay"
		menu.addItem vrayShowVFB -1
	)
)

macroScript VRayProperties category:"VRay" buttontext:"V-Ray properties" tooltip:"Displays the VRay object or light properties" (
	doVRayObjectProperties()
)

macroScript VRaySceneConverter category:"VRay" buttontext:"V-Ray scene converter" tooltip:"Convert all scene materials to VRayMtl and all raytraced shadows to VRay shadows" (
	CreateDialog vraySceneConverterRollout modal:true
)

macroScript VRayBitmap2VRayHDRITexmapConverter
	category:"VRay"
	buttonText:"V-Ray Bitmap to VRayHDRI converter"
	tooltip:"Convert all scene materials' Bitmaps to VRayHDRI"
(
	CreateDialog convertB2HDRI_Rollout modal:true
)

macroScript VRayMeshExport category:"VRay" buttontext:"V-Ray mesh export" tooltip:"Exports the selected meshes to an external file" (
	doVRayMeshExport()
)

macroScript VRayShowVFB category:"VRay" buttontext:"V-Ray VFB" tooltip:"Displays the last rendered V-Ray VFB, if is exists" (
	r=renderers.current
	cls=r.classid
	if (cls[1]==1941615238 and cls[2]==2012806412) then r.showLastVFB()
)

-- Menu bar V-Ray action items
macroScript VRayProxy
	category:"VRay"
	buttonText:"V-Ray Proxy"
	toolTip: "Create V-Ray Proxy"
(
	on execute do StartObjectCreation VRayProxy
)

macroScript VRayFur
	category:"VRay"
	buttonText:"V-Ray Fur"
	toolTip: "Create V-Ray Fur"
(
	fn setSourceNode obj = obj.sourceNode = $
	
	on execute do startObjectCreation Vrayfur newNodeCallback: setSourceNode
	on isEnabled do (superClassOf $ == GeometryClass)
)

macroScript VRayPlane
	category:"VRay"
	buttonText:"V-Ray Plane"
	toolTip: "Create V-Ray Plane"
(
	on execute do StartObjectCreation VRayPlane
)

macroScript VRaySphere
	category:"VRay"
	buttonText:"V-Ray Sphere"
	toolTip: "Create V-Ray Sphere"
(
	on execute do StartObjectCreation VRaySphere
)

macroScript VRayLight_Plane
	category:"VRay"
	buttonText:"V-Ray Plane Light"
	toolTip: "Create V-Ray Plane Light"
(
	fn setType obj = (obj.type = 0)
	
	on execute do StartObjectCreation VRayLight newNodeCallback: setType
)

macroScript VRayLight_Dome
	category:"VRay"
	buttonText:"V-Ray Dome Light"
	toolTip: "Create V-Ray Dome Light"
(
	fn setType obj = (obj.type = 1)
	
	on execute do StartObjectCreation VRayLight newNodeCallback: setType
)

macroScript VRayLight_Sphere
	category:"VRay"
	buttonText:"V-Ray Sphere Light"
	toolTip: "Create V-Ray Sphere Light"
(
	fn setType obj = (obj.type = 2)
	
	on execute do StartObjectCreation VRayLight newNodeCallback: setType
)

macroScript VRayLight_Mesh
	category:"VRay"
	buttonText:"V-Ray Mesh Light"
	toolTip: "Create V-Ray Mesh Light"
(
	fn setType obj = (obj.type = 3)
	
	on execute do StartObjectCreation VRayLight newNodeCallback: setType
)

macroScript VRayIES
	category:"VRay"
	buttonText:"V-Ray IES Light"
	toolTip: "Create V-Ray IES Light"
(
	on execute do StartObjectCreation VRayIES
)

macroScript VRayAmbientLight
	category:"VRay"
	buttonText:"V-Ray Ambient Light"
	toolTip: "Create V-Ray Ambient Light"
(
	on execute do StartObjectCreation VRayAmbientLight
)

macroScript VRaySun
	category:"VRay"
	buttonText:"V-Ray Sun"
	toolTip: "Create V-Ray Sun"
(
	on execute do StartObjectCreation VRaySun
)

macroScript VRayDomeCamera
	category:"VRay"
	buttonText:"V-Ray Dome Camera"
	toolTip: "Create V-Ray Dome Camera"
(
	on execute do StartObjectCreation VRayDomeCamera
)

macroScript VRayPhysicalCamera
	category:"VRay"
	buttonText:"V-Ray Physical Camera"
	toolTip: "Create V-Ray Physical Camera"
(
	on execute do StartObjectCreation VRayPhysicalCamera
)

macroScript VRayLightMeter
	category:"VRay"
	buttonText:"V-Ray Light Meter"
	toolTip: "Create V-Ray Light Meter"
(
	on execute do StartObjectCreation VRayLightMeter
)

macroscript VRayDisplacementMod
	category:"VRay"
	buttonText:"V-Ray Displacement Mod"
	toolTip: "V-Ray Displacement Mod"
(
	on execute do AddMod VRayDisplacementMod
	on isEnabled return ValidModifier $ VRayDisplacementMod
)

macroscript VRayOrnatrixMod
	category:"VRay"
	buttonText:"V-Ray Ornatrix Mod"
	toolTip: "V-Ray Ornatrix Mod"
(
	on execute do AddMod VRayOrnatrixMod
	on isEnabled return ValidModifier $ VRayOrnatrixMod
)

macroScript VRayRTAnimPreview
enabledIn:#("max")
category:"VRay"
buttontext:"Make V-Ray RT Animation Preview"
tooltip:"Render an animation preview with V-Ray RT"
(
	local enabled = false
	local running = false
	local asRenderer = undefined
	local cFrame
	local saveFilePath = (GetDir #renderoutput) + "\frame"
	local saveFileExt = ".png"
	local dialogPos
	local startFrame = (animationRange.start as Integer)/160
	local endFrame = (animationRange.end as Integer)/160
	
	rollout SettingsRollout "V-Ray RT Animation Preview Render Settings" width:262 height:216 (
		label startFrameSpinLabel "Start frame:" pos:[16, 10] width:160 height:16
		spinner startFrameSpin "" pos:[175, 10] width:70 height:16 range:[0, animationRange.end, startFrame] type:#integer scale:1
		
		label endFrameSpinLabel "End frame:" pos:[16, 32] width:160 height:16
		spinner endFrameSpin "" pos:[175, 32] width:70 height:16 range:[0, animationRange.end, endFrame] type:#integer scale:1
		
		label timeLimitSpinLabel "Time limit (min.):" pos:[16, 54] width:160 height:16
		spinner timeLimitSpin "" pos:[175, 54] width:70 height:16 range:[0.0, 525600.0, asRenderer.max_render_time] type:#float scale:0.25
		
		label noiseThrLimitSpinLabel "Noise threshold:" pos:[16, 76] width:160 height:16
		spinner noiseThrLimitSpin "" pos:[175, 76] width:70 height:16 range:[0.00001, 0.99, asRenderer.aa_threshold] type:#float scale:0.01
		
		label maxPathsLimitSpinLabel "Max. paths per pixel:" pos:[16, 98] width:160 height:16
		spinner maxPathsLimitSpin "" pos:[175, 98] width:70 height:16 range:[0, 65536, asRenderer.max_paths_per_pixel] type:#integer scale:1
		
		edittext saveFileNameEdit "" text:(saveFilePath + saveFileExt) pos:[12, 120] width:208 height:17
		button saveFileNameBtn "..." pos:[226, 120] width:20 height:18

		button startPreviewBtn "Start Preview Render" pos:[16, 142] width:230 height:18
		
		button cancelPreview "Cancel" pos:[16, 164] width:230 height:18

		progressBar animProgress pos:[16, 186] width:230 height:16
		
		on startFrameSpin changed val do (
			startFrame = val
		)
		
		on endFrameSpin changed val do (
			endFrame = val
		)
		
		on timeLimitSpin changed val do (
			if (enabled) do (
				asRenderer.max_render_time = val
			)
		)
		
		on noiseThrLimitSpin changed val do (
			if (enabled) do (
				asRenderer.aa_threshold = val
			)
		)

		on maxPathsLimitSpin changed val do (
			if (enabled) do (
				asRenderer.max_paths_per_pixel = val
			)
		)
		
		on saveFileNameEdit changed val do (
			saveFilePath = (getFilenamePath val) + (getFilenameFile val)
			saveFileExt = (getFilenameType val)
		)
		
		on saveFileNameBtn pressed do (
			local newFilePath = getSaveFileName filename:saveFileNameEdit.text \
											types:"PNG Image File (*.png)|*.png|JPEG File (*.jpg)|*.jpg" \
											historyCategory:"MAXScriptFileOpenSave"
			if (newFilePath != undefined) do (
				saveFileNameEdit.text = newFilePath
				saveFilePath = (getFilenamePath newFilePath) + (getFilenameFile newFilePath)
				saveFileExt = (getFilenameType newFilePath)
			)
		)
		
		on startPreviewBtn pressed do (
			if (running == false) do (
				running = true
				startPreviewBtn.enabled = false
				animProgress.value = 0
				cFrame = startFrame
				saveFilePath = (getFilenamePath saveFileNameEdit.text) + (getFilenameFile saveFileNameEdit.text)
				saveFileExt = (getFilenameType saveFileNameEdit.text)
				
				if (vrayGetRTBitmap() != undefined) then (
					sliderTime = cFrame
					actionMan.executeAction 369690881 "1"	-- reinitialize renderer
				)
				else (
					sliderTime = cFrame
					actionMan.executeAction 0 "40701"	-- start VrayRT ActiveShade
				)
			)
		)
		
		on cancelPreview pressed do (
			startPreviewBtn.enabled = true
			running = false
		)
		
		on SettingsRollout open do (
			if (startFrame < animationRange.start) do (
				startFrame = animationRange.start
			)
			
			if (endFrame > animationRange.end) do (
				endFrame = animationRange.end
			)
		)
		
		on SettingsRollout moved pos do (
			dialogPos = pos
		)
		
		on SettingsRollout close do (
			running = false
		)
	)
	
	fn checkRenderer = (
		asRenderer = renderers.activeShade
		
		if (asRenderer != undefined) then (
			enabled = ((renderers.activeShade.classid[1] == 1770671000) and (renderers.activeShade.classid[2] == 1323107829))
		) else (
			enabled = false
		)
	)
	
	on execute do (
		checkRenderer()
		
		if (enabled) do (
			if (dialogPos != undefined) then (
				CreateDialog SettingsRollout style:#(#style_titlebar, #style_border, #style_sysmenu) pos:dialogPos modal:true
			) else (
				CreateDialog SettingsRollout style:#(#style_titlebar, #style_border, #style_sysmenu) modal:true
			)
		)
	)
	
	on isEnabled do (
		checkRenderer()
		enabled
	)
	
	on isVisible do (
		checkRenderer()
		enabled
	)

	-- Export function through global variable declaration
	global vrayRTImageComplete
	fn vrayRTImageComplete reason = (
		if (running) do (
			if ((saveFilePath != "") AND (saveFileExt != "")) do (
				local b = vrayGetRTBitmap()
				if (b != undefined) do (
					b.filename = (saveFilePath + (formattedPrint (cFrame) format:"04i") + saveFileExt)
					save b
					close b
				)
			)
			
			SettingsRollout.animProgress.value = ((((cFrame - startFrame)+1)*100)/((endFrame - startFrame)+1))
			
			if (cFrame < endFrame) then (
				cFrame += 1
				sliderTime = cFrame
				actionMan.executeAction 369690881 "1"	-- reinitialize renderer
			) else (
				SettingsRollout.startPreviewBtn.enabled = true
				running = false
			)
		)
	)
)

-- Localized menu names
include "$startupScripts\\vrayutils_local.ms"

-- Menu editing utility functions
-- Append menu item to a menu object;
-- Parameters:	Menu
--		Item's MacroScript Name,
--		Item's MacroScript Category,
fn appendMenuItem mobj msname mscat = mobj.addItem (menuMan.createActionItem msname mscat) -1

-- Insert menu item to a menu object at specified position;
-- Parameters:	Menu
--		Item's MacroScript Name,
--		Item's MacroScript Category,
-- 		Item's insert position
fn InsertMenuItem mobj msname mscat idx = mobj.addItem (menuMan.createActionItem msname mscat) idx

-- Insert submenu item:
-- Parameters:	Menu
--		Menu Name
--		Submenu Object
--		Menu Index
fn insertSubmenuItem mobj mname sobj mindex = (
	local subMenuItem = menuMan.createSubMenuItem mname sobj
	subMenuItem.setUseCustomTitle true
	mobj.addItem subMenuItem mindex
)

-- Return index of the first separator in a menu, if no separator last element's index is returned
-- Parameters:	Menu
fn findFirstSeparatorIdx mobj = (
	local i = 1
	while ((i <= mobj.numItems()) and (not((mobj.getItem i).getIsSeparator()))) do (
		i += 1
	)
	
	i
)

-- Return menu item from a menu, found by title
-- Parameters:	Menu
--		Item Name
fn findMenuItem mobj iname = (
	local item = undefined
	local i
	for i = 1 to mobj.numItems() do (
		local titleElements = filterString ((mobj.getItem i).getTitle()) "&"
		local title = ""
		local j
		for j = 1 to titleElements.count do ( 
			title = append title titleElements[j]
		)
		
		if (title == iname) do (
			item = mobj.getItem i
		)
	)
	
	item
)

-- Return menu item's index from a menu, found by title
-- Parameters:	Menu
--		Item Name
fn findMenuItemIdx mobj iname = (
	local item = undefined
	local i
	for i = 1 to mobj.numItems() do (
		local titleElements = filterString ((mobj.getItem i).getTitle()) "&"
		local title = ""
		local j
		for j = 1 to titleElements.count do ( 
			title = append title titleElements[j]
		)
		
		if (title == iname) do (
			item = i
		)
	)
	
	item
)

-- Register V-Ray objects in menu bar
fn registerVRayMenuBarItems = (
	local updateEnable = false
	local subMenu -- temporary submenu object
	local menuIndex -- temporary index used for item positioning calculation
	
	local mainMenuBar = menuMan.getMainMenuBar()
	
	local toolsMenuItem
	for i=1 to menuName_Tools.count do (if (toolsMenuItem==undefined) do (toolsMenuItem=findMenuItem mainMenuBar menuName_Tools[i]))
	if (toolsMenuItem != undefined) do (
		local toolsMenu = toolsMenuItem.getSubMenu()
		
		-- Add Light Lister item
		if ((findMenuItem toolsMenu "V-Ray Light Lister") == undefined) do (
			local lightListerIndex
			for i=1 to menuName_Tools_LightLister.count do (if (lightListerIndex==undefined) do (lightListerIndex=findMenuItemIdx toolsMenu menuName_Tools_LightLister[i]))
			if (lightListerIndex != undefined) then (
				lightListerIndex += 1
			)
			else (
				lightListerIndex = -1
			)
			toolsMenu.addItem (menuMan.createActionItem "VRayLightLister" "VRay") lightListerIndex
			
			updateEnable = true
		)
	)
	
	local createMenuItem
	for i=1 to menuName_Create.count do (if (createMenuItem==undefined) do (createMenuItem=findMenuItem mainMenuBar menuName_Create[i]))
	if (createMenuItem != undefined) do (
		local createMenu = createMenuItem.getSubMenu()
		
		-- Add geometry objects
		if ((findMenuItem createMenu "V-Ray") == undefined) do (
			subMenu = menuMan.createMenu "Create - V-Ray Primitives"
			-- Add items to the submenu
			appendMenuItem subMenu "VRayProxy" "VRay"
			appendMenuItem subMenu "VRayFur" "VRay"
			appendMenuItem subMenu "VRayPlane" "VRay"
			appendMenuItem subMenu "VRaySphere" "VRay"
			
			-- Add submenu before the first separator
			menuIndex = findFirstSeparatorIdx createMenu
			insertSubmenuItem createMenu "V-Ray" subMenu menuIndex
			updateEnable = true
		)
		
		-- Add light objects
		local lightsSubMenuItem
		for i=1 to menuName_Create_Lights.count do (if (lightsSubMenuItem==undefined) do (lightsSubMenuItem=findMenuItem createMenu menuName_Create_Lights[i]))
		if (lightsSubMenuItem != undefined) do (
			local lightsSubMenu = lightsSubMenuItem.getSubMenu()
			if ((findMenuItem lightsSubMenu "V-Ray") == undefined) do (
				subMenu = menuMan.createMenu "Create - Lights.V-Ray Lights"
				
				-- Add items to the submenu
				appendMenuItem subMenu "VRayLight_Plane" "VRay"
				appendMenuItem subMenu "VRayLight_Dome" "VRay"
				appendMenuItem subMenu "VRayLight_Sphere" "VRay"
				appendMenuItem subMenu "VRayLight_Mesh" "VRay"
				appendMenuItem subMenu "VRayIES" "VRay"
				appendMenuItem subMenu "VRayAmbientLight" "VRay"
				appendMenuItem subMenu "VRaySun" "VRay"
				
				-- Add submenu before the first separator
				menuIndex = findFirstSeparatorIdx lightsSubMenu
				insertSubmenuItem lightsSubMenu "V-Ray" subMenu menuIndex
				updateEnable = true
			)
		)
		
		
		-- Add camera objects
		local camerasSubMenuItem
		for i=1 to menuName_Create_Cameras.count do (if (camerasSubMenuItem==undefined) do (camerasSubMenuItem=findMenuItem createMenu menuName_Create_Cameras[i]))
		if (camerasSubMenuItem != undefined) do (
			local camerasSubMenu = camerasSubMenuItem.getSubMenu()
			if ((findMenuItem camerasSubMenu "V-Ray") == undefined) do (
				subMenu = menuMan.createMenu "Create - Cameras.V-Ray Cameras"
				
				-- Add items to the submenu
				appendMenuItem subMenu "VRayDomeCamera" "VRay"
				appendMenuItem subMenu "VRayPhysicalCamera" "VRay"
				
				-- Add submenu before the first separator
				menuIndex = findFirstSeparatorIdx camerasSubMenu
				insertSubmenuItem camerasSubMenu "V-Ray" subMenu menuIndex
				updateEnable = true
			)
		)
		
		
		-- Add helper objects
		local helpersSubMenuItem
		for i=1 to menuName_Create_Helpers.count do (if (helpersSubMenuItem==undefined) do (helpersSubMenuItem=findMenuItem createMenu menuName_Create_Helpers[i]))
		if (helpersSubMenuItem != undefined) do (
			local helpersSubMenu = helpersSubMenuItem.getSubMenu()
			if ((findMenuItem helpersSubMenu "V-Ray") == undefined) do (
				subMenu = menuMan.createMenu "Create - Helpers.V-Ray Helpers"
				
				-- Add items to the submenu
				appendMenuItem subMenu "VRayLightMeter" "VRay"
				
				-- Add submenu at the end
				helpersSubMenu.addItem (menuMan.createSeparatorItem()) -1
				insertSubmenuItem helpersSubMenu "V-Ray" subMenu -1
				updateEnable = true
			)
		)
	)
	
	local modMenuItem
	for i=1 to menuName_Modifiers.count do (if (modMenuItem==undefined) do (modMenuItem=findMenuItem mainMenuBar menuName_Modifiers[i]))
	if (modMenuItem != undefined) do (
		local modMenu = modMenuItem.getSubMenu()
		
		-- Add modifiers
		if ((findMenuItem modMenu "V-Ray") == undefined) do (
			subMenu = menuMan.createMenu "Modifiers - V-Ray"
			-- Add items to the submenu
			appendMenuItem subMenu "VRayDisplacementMod" "VRay"
			appendMenuItem subMenu "VRayOrnatrixMod" "VRay"
			
			modMenu.addItem (menuMan.createSeparatorItem()) -1
			insertSubmenuItem modMenu "V-Ray" subMenu -1
			updateEnable = true
		)
	)
	
	local renderingMenuItem = undefined
	for i=1 to mainMenuBar_Rendering.count do (if (renderingMenuItem==undefined) do (renderingMenuItem=findMenuItem mainMenuBar mainMenuBar_Rendering[i]))
	if (renderingMenuItem != undefined) do (
		local renderingMenu = renderingMenuItem.getSubMenu()
		
		if ((findMenuItem renderingMenu "Make V-Ray RT Animation Preview") == undefined) do (
			-- Add item
			renderingMenu.addItem (menuMan.createSeparatorItem()) -1
			appendMenuItem renderingMenu "VRayRTAnimPreview" "VRay"
			
			updateEnable = true
		)
	)
	
	-- Update menu to show all newly added items
	if (updateEnable == true) do (
		menuMan.updateMenuBar()
	)
)

fn registerVRayQuadMenuItems = (
	local quadMenu=menuMan.getViewportRightClickMenu #nonePressed
	if (quadMenu!=undefined) do (
		local menu=quadMenu.getMenu 1
	)
	
	local menuIndex
	
	if (menu!=undefined) do (
		-- add Bitmap2VRayHDRI texture converter
		menuIndex=findMenuItemIdx menu "V-Ray scene converter"
		appendMenuItem menu "VRayBitmap2VRayHDRITexmapConverter" "VRay"
	)
)

-- Unregister V-Ray objects from menu bar
fn unregisterVRayMenuBarItems = (
	local mainMenuBar = menuMan.getMainMenuBar()
	local vrayMenuItemIdx

	local updateEnable = false

	local toolsMenuItem
	for i=1 to menuName_Tools.count do (if (toolsMenuItem==undefined) do (toolsMenuItem=findMenuItem mainMenuBar menuName_Tools[i]))
	if (toolsMenuItem != undefined) do (
		local toolsMenu = toolsMenuItem.getSubMenu()
		if (toolsMenu != undefined) do (
			vrayMenuItemIdx = findMenuItemIdx toolsMenu "V-Ray Light Lister"
			if (vrayMenuItemIdx != undefined) do (
				toolsMenu.removeItemByPosition (vrayMenuItemIdx)
			)
		)
	)

	local createMenuItem
	for i=1 to menuName_Create.count do (if (createMenuItem==undefined) do (createMenuItem=findMenuItem mainMenuBar menuName_Create[i]))
	if(createMenuItem != undefined) do (
		local createMenu = createMenuItem.getSubMenu()
		local helpersMenuItem
		for i=1 to menuName_Create_Helpers.count do (if (helpersMenuItem==undefined) do (helpersMenuItem=findMenuItem createMenu menuName_Create_Helpers[i]))
		if(helpersMenuItem != undefined) do (
			local helpersMenu = helpersMenuItem.getSubMenu()
			vrayMenuItemIdx = findMenuItemIdx helpersMenu "V-Ray"
			if ((vrayMenuItemIdx != undefined) and (vrayMenuItemIdx > 1)) do (
				if((helpersMenu.getItem (vrayMenuItemIdx-1)).getIsSeparator()) do (
					helpersMenu.removeItemByPosition (vrayMenuItemIdx-1)
					updateEnable = true
				)
			)
		)
	)
	
	local modMenuItem
	for i=1 to menuName_Modifiers.count do (if (modMenuItem==undefined) do (modMenuItem=findMenuItem mainMenuBar menuName_Modifiers[i]))
	if (modMenuItem != undefined) do (
		local modMenu = modMenuItem.getSubMenu()
		vrayMenuItemIdx = findMenuItemIdx modMenu "V-Ray"
		if ((vrayMenuItemIdx != undefined) and (vrayMenuItemIdx > 1)) do (
			if ((modMenu.getItem (vrayMenuItemIdx-1)).getIsSeparator()) do (
				modMenu.removeItemByPosition (vrayMenuItemIdx-1)
				updateEnable = true
			)
		)
	)
	
	local renderingMenuItem = undefined
	for i=1 to mainMenuBar_Rendering.count do (if (renderingMenuItem==undefined) do (renderingMenuItem=findMenuItem mainMenuBar mainMenuBar_Rendering[i]))
	if (renderingMenuItem != undefined) do (
		local renderingMenu = renderingMenuItem.getSubMenu()
		vrayMenuItemIdx = findMenuItemIdx renderingMenu "Make V-Ray RT Animation Preview"
		if (vrayMenuItemIdx != undefined) do (
			renderingMenu.removeItemByPosition vrayMenuItemIdx
			if (vrayMenuItemIdx > 1) do (
				if ((renderingMenu.getItem (vrayMenuItemIdx-1)).getIsSeparator()) do (
					renderingMenu.removeItemByPosition (vrayMenuItemIdx-1)
				)
			)
			updateEnable = true
		)
	)
	
	-- unregister all V-Ray submenus
	local vrayMenu
	vrayMenu = menuMan.findMenu "Create - V-Ray Primitives"
	if (vrayMenu != undefined) do (
		menuMan.unregisterMenu vrayMenu
		updateEnable = true
	)

	vrayMenu = menuMan.findMenu "Create - Lights.V-Ray Lights"
	if (vrayMenu != undefined) do (
		menuMan.unregisterMenu vrayMenu
		updateEnable = true
	)

	vrayMenu = menuMan.findMenu "Create - Cameras.V-Ray Cameras"
	if (vrayMenu != undefined) do (
		menuMan.unregisterMenu vrayMenu
		updateEnable = true
	)

	vrayMenu = menuMan.findMenu "Create - Helpers.V-Ray Helpers"
	if (vrayMenu != undefined) do (
		menuMan.unregisterMenu vrayMenu
		updateEnable = true
	)

	vrayMenu = menuMan.findMenu "Modifiers - V-Ray"
	if (vrayMenu != undefined) do (
		menuMan.unregisterMenu vrayMenu
		updateEnable = true
	)


	-- Update menu to show all newly added items
	if (updateEnable == true) do (
		menuMan.updateMenuBar()
	)
)

fn unregisterVRayQuadMenuItems = (
	local quadMenu=menuMan.getViewportRightClickMenu #nonePressed
	if (quadMenu!=undefined) do (
		local menu=quadMenu.getMenu 1
	)
	
	local menuIndex
	
	if (menu!=undefined) do (
		-- remove Bitmap2VRayHDRI texture converter
		menuIndex=findMenuItemIdx menu "V-Ray Bitmap to VRayHDRI converter"
		if (menuIndex!=undefined) do (
			menu.removeItemByPosition menuIndex
		)
	)
)

fn registerVRayMenus = (
	registerVRayMenuBarItems()
	registerVRayQuadMenuItems()
	
	if (registerVRayRTMenus != undefined) do (registerVRayRTMenus())
	
	menuMan.saveMenuFile (menuMan.getMenuFile())
)

fn unregisterVRayMenus = (
	unregisterVRayMenuBarItems()
	unregisterVRayQuadMenuItems()
)

fn updateVRayMenus = (
	unregisterVRayMenus()
	registerVRayMenus()
)

fn setupVRayUI = (
	registerMenuItems()
)

setupVRayUI()